@@ -74,17 +74,20 @@ sub _create_tcp_server {
         $port = $listen;
     }
 
-    return tcp_server $host, $port, $self->_accept_handler($app, $is_tcp),
-        $self->_accept_prepare_handler;
+    my($listen_host, $listen_port);
+
+    return tcp_server $host, $port, $self->_accept_handler($app, $is_tcp, \$listen_host, \$listen_port),
+        $self->_accept_prepare_handler(\$listen_host, \$listen_port);
 }
 
 sub _accept_prepare_handler {
-    my $self = shift;
+    my($self, $listen_host_r, $listen_port_r) = @_;
+
     return sub {
         my ( $fh, $host, $port ) = @_;
         DEBUG && warn "Listening on $host:$port\n";
-        $self->{prepared_host} = $host;
-        $self->{prepared_port} = $port;
+        $$listen_host_r = $host;
+        $$listen_port_r = $port;
         $self->{server_ready}->({
             host => $host,
             port => $port,
@@ -96,7 +99,7 @@ sub _accept_prepare_handler {
 }
 
 sub _accept_handler {
-    my ( $self, $app, $is_tcp ) = @_;
+    my ( $self, $app, $is_tcp, $listen_host_r, $listen_port_r ) = @_;
 
     return sub {
         my ( $sock, $peer_host, $peer_port ) = @_;
@@ -114,8 +117,8 @@ sub _accept_handler {
         my $try_parse = sub {
             if ( $self->_try_read_headers($sock, $headers) ) {
                 my $env = {
-                    SERVER_PORT         => $self->{prepared_port},
-                    SERVER_NAME         => $self->{prepared_host},
+                    SERVER_NAME         => $$listen_host_r,
+                    SERVER_PORT         => $$listen_port_r,
                     SCRIPT_NAME         => '',
                     REMOTE_ADDR         => $peer_host,
                     'psgi.version'      => [ 1, 0 ],
@@ -157,7 +160,8 @@ sub _accept_handler {
 
             1;
         }) {
-            $self->_bad_request($sock);
+            my $disconnected = ($@ =~ /^client disconnected/);
+            $self->_bad_request($sock, $disconnected);
         }
     };
 }
@@ -214,24 +218,27 @@ sub _create_req_parsing_watcher {
         } catch {
             undef $headers_io_watcher;
             undef $timeout_timer;
-            $self->_bad_request($sock);
+            my $disconnected = /^client disconnected/;
+            $self->_bad_request($sock, $disconnected);
         }
     };
 }
 
 sub _bad_request {
-    my ( $self, $sock ) = @_;
+    my ( $self, $sock, $disconnected ) = @_;
 
     return unless defined $sock and defined fileno $sock;
 
-    $self->_write_psgi_response(
-        $sock,
-        [
-            400,
-            [ 'Content-Type' => 'text/plain' ],
-            [ ],
-        ],
-    );
+    my $response = [
+        400,
+        [ 'Content-Type' => 'text/plain' ],
+        [ ],
+    ];
+
+    # if client is already gone, don't try to write to it
+    $response = [] if $disconnected;
+
+    $self->_write_psgi_response($sock, $response);
 
     return;
 }
@@ -261,6 +268,8 @@ sub _read_chunk {
                 return 1;
             } elsif ($! and $! != EAGAIN && $! != EINTR && $! != WSAEWOULDBLOCK) {
                 die $!;
+            } elsif (!$!) {
+                die "client disconnected";
             }
         }
 
@@ -285,7 +294,7 @@ sub _run_app {
     my($self, $app, $env, $sock) = @_;
 
     unless ($env->{'psgi.input'}) {
-        if ($env->{CONTENT_LENGTH} && $env->{REQUEST_METHOD} =~ /^(?:POST|PUT)$/) {
+        if ($env->{CONTENT_LENGTH}) {
             $self->_read_chunk($sock, $env->{CONTENT_LENGTH}, sub {
                 my ($data) = @_;
                 open my $input, '<', \$data;
@@ -303,6 +312,7 @@ sub _run_app {
     if ( ref $res eq 'ARRAY' ) {
         $self->_write_psgi_response($sock, $res);
     } elsif ( blessed($res) and $res->isa("AnyEvent::CondVar") ) {
+        Carp::carp("Returning AnyEvent condvar is deprecated and will be removed in the next release of Twiggy. Use the streaming callback interface intstead.");
         $res->cb(sub { $self->_write_psgi_response($sock, shift->recv) });
     } elsif ( ref $res eq 'CODE' ) {
         $res->(
